#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "chess.h"
#include "data.h"
#include "epdglue.h"

/* modified 03/31/98 */
/*
********************************************************************************
*                                                                              *
*   Search() is the recursive routine used to implement the alpha/beta         *
*   negamax search (similar to minimax but simpler to code.)  Search() is      *
*   called whenever there is "depth" remaining so that all moves are subject   *
*   to searching, or when the side to move is in check, to make sure that this *
*   side isn't mated.  Search() recursively calls itself until depth is ex-    *
*   hausted, at which time it calls Quiesce() instead.                         *
*                                                                              *
********************************************************************************
*/
int Thread(TREE *tree, int alpha, int beta, int value, int wtm, int depth,
           int ply, int threat) {
  SMP_PARMS parms[CPUS];
  thread_t pthreads[CPUS];
  TREE *block;
  int i, proc, nblocks=0;
/*
 ----------------------------------------------------------
|                                                          |
|   first, we copy the data from the parent thread to each |
|   child process's memory area.                           |
|                                                          |
 ----------------------------------------------------------
*/
  Lock(lock_smp);
  if (smp_busy) {
    UnLock(lock_smp);
    return(0);
  }
  smp_busy=1;
# if defined(DEBUGSMP)
  Print(128,"Parallel search split using %d processors at ply %d\n",max_threads,ply);
#endif
/*
 ----------------------------------------------------------
|                                                          |
|   now pack the parameters to SearchSMP() into a          |
|   structure to be passed thru the thread create call.    |
|   note that we also make a list of each "tree" block we  |
|   allocate here, and then copy those into the "sibling"  |
|   vector of each thread's tree block.  the "parent"      |
|   entry in the "tree" block points to the shared parent  |
|   tree block, so this set of pointers links a process to |
|   its parent, and to its siblings.                       |
|                                                          |
 ----------------------------------------------------------
*/
  for (proc=0;proc<max_threads;proc++) {
    block=CopyToSMP(tree,ply);
    if (!block) break;
    nblocks++;
    tree->siblings[proc]=block;
    parms[proc].tree=block;
    parms[proc].shared=tree;
    parms[proc].alpha=alpha;
    parms[proc].beta=beta;
    parms[proc].value=value;
    parms[proc].wtm=wtm;
    parms[proc].depth=depth;
    parms[proc].ply=ply;
    parms[proc].threat=threat;
    parms[proc].cpu_id='a'+proc;
    block->stop=0;
    block->parent=tree;
    block->nprocs=max_threads;
    for (i=0;i<CPUS;i++) block->siblings[i]=0;
  }
  tree->search_value=value;
  UnLock(lock_smp);
  if (!nblocks) return(0);
/*
 ----------------------------------------------------------
|                                                          |
|   now create max_threads threads to work together to     |
|   search the current node.                               |
|                                                          |
 ----------------------------------------------------------
*/
  for (proc=0;proc<nblocks;proc++)
    tfork(pthreads[proc],SearchSMP,parms[proc]);
/*
 ----------------------------------------------------------
|                                                          |
|   after the threads are kicked off to start the parallel |
|   search, we then wait for each to complete.  after they |
|   finish, we compare their result to the best so far and |
|   overwrite the best if needed.                          |
|                                                          |
 ----------------------------------------------------------
*/
  for (proc=0;proc<nblocks;proc++)
    tjoin(pthreads[proc]);
  for (proc=0;proc<nblocks;proc++)
    CopyFromSMP(tree,tree->siblings[proc],ply);
# if defined(DEBUGSMP)
  Print(128,"Parallel join at ply %d\n",ply);
#endif
  smp_busy=0;
  return(1);
}
