#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "chess.h"
#include "data.h"
#define THREAT_MARGIN 150

/*
********************************************************************************
*                                                                              *
*   Threat() is used to extend the depth for positions that appear to have     *
*   some threat.  the basic algorithm used is a null move search, but with the *
*   alpha/beta window set to (v-n,v-n+1) where n is some significant score     *
*   (normally 1.5 pawns).  Threat() is called when search returns a value that *
*   is > alpha (>= beta means this is a fail-hi node, while > alpha < beta     *
*   means this is a PV node.)  Threat() simply does a null move search with    *
*   the sharply lowered bounds (given above) to see if this move is preventing *
*   the score from dropping dramatically (such as happens for horizon effect   *
*   moves.)  if the proposed move fails hi, but "passing" fails low by a large *
*   amount, then something is "fishy" and the position needs to be searched    *
*   again, but with an increased search depth to try and resolve what's going  *
*   on.                                                                        *
*                                                                              *
********************************************************************************
*/
int Threat(int value, int alpha, int beta, int wtm, int depth, int ply)
{
  BITBOARD save_hash_key;
  register int new_value, temp, tnodes;
  CHESS_PATH temp_p;
/*
 ----------------------------------------------------------
|                                                          |
|   if the current ply has already been extended by this   |
|   algorithm, we just research the current fail-high move |
|   to depth+1, the null-move threat test is not necessary |
|   again.                                                 |
|                                                          |
 ----------------------------------------------------------
*/
  if (!(extended_reason[ply]&threat_extension)) {
/*
 ----------------------------------------------------------
|                                                          |
|   if the remaining full-width depth is not at least 4,   |
|   then don't try threat extensions.                      |
|                                                          |
 ----------------------------------------------------------
*/
    if (depth < 4*INCREMENT_PLY) return(value);
/*
 ----------------------------------------------------------
|                                                          |
|   if the remaining search has already been extended an   |
|   excessive amount, don't do threat extensions too.      |
|                                                          |
 ----------------------------------------------------------
*/
    if (ply > 3*iteration_depth) return(value);
/*
 ----------------------------------------------------------
|                                                          |
|   if the side on move is winning significantly, then the |
|   test is worthless and should be skipped.               |
|                                                          |
 ----------------------------------------------------------
*/
    if (( wtm && Material-THREAT_MARGIN > beta) ||
        (!wtm && -Material-THREAT_MARGIN > beta))
      return(value);
/*
 ----------------------------------------------------------
|                                                          |
|   if the current move is capturing a piece moved at the  |
|   previous ply, then the threat extension will probably  |
|   fail low, since we would then not be re-capturing this |
|   piece and let it slip away.                            |
|                                                          |
 ----------------------------------------------------------
*/
    if (Captured(current_move[ply]) &&
        (To(current_move[ply]) == To(current_move[ply-1])))
      return(value);
/*
 ----------------------------------------------------------
|                                                          |
|   first, do a null-move search with the lowered window   |
|   to see if this fails low or high.                      |
|                                                          |
 ----------------------------------------------------------
*/
    tnodes=nodes_searched;
    temp=current_move[ply];
    UnMakeMove(ply,current_move[ply],wtm);
    temp_p=pv[ply];
    current_move[ply]=0;
    current_phase[ply]=NULL_MOVE;
#if !defined(FAST)
    if (ply <= trace_level)
      SearchTrace(ply,depth,wtm,alpha,beta,"ThreatTest",0);
#endif
    position[ply+1]=position[ply];
    Rule50Moves(ply+1)++;
    save_hash_key=HashKey;
    if (EnPassant(ply)) {
      HashEP(EnPassant(ply+1),HashKey);
      EnPassant(ply+1)=0;
    }
    new_value=-Search(-value+THREAT_MARGIN-1, -value+THREAT_MARGIN,
                      ChangeSide(wtm),depth-NULL_MOVE_DEPTH-INCREMENT_PLY,ply+1,0);
    HashKey=save_hash_key;
    current_move[ply]=temp;
    MakeMove(ply,current_move[ply],wtm);
    pv[ply]=temp_p;
    t_nodes_searched+=nodes_searched-tnodes;
    if (abort_search) return(0);
    if (new_value > value-THREAT_MARGIN) return(value);
    extended_reason[ply]|=threat_extension;
  }
/*
 ----------------------------------------------------------
|                                                          |
|   if the null-move searched failed low, then something   |
|   is wrong with this move/position.  search it again,    |
|   but one ply deeper than before.                        |
|                                                          |
 ----------------------------------------------------------
*/
  threat_extensions_done++;
  UnMakeMove(ply,current_move[ply],wtm);
#if !defined(FAST)
  if (ply <= trace_level)
    SearchTrace(ply,depth,wtm,alpha,beta,"ThreatMove",current_phase[ply]);
#endif
  MakeMove(ply,current_move[ply],wtm);
  value=-Search(-beta,-alpha,ChangeSide(wtm),depth,ply+1,1);
  if (abort_search) return(0);
  return(value);
}
